<?php

namespace Drupal\commerce_gnu_taler\Plugin\Commerce\PaymentGateway;

use Drupal\commerce\MailHandlerInterface;
use Drupal\commerce_gnu_taler\OrderAdapter;
use Drupal\commerce_gnu_taler\PluginForm\PaymentForm;
use Drupal\commerce_gnu_taler\TalerClient;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Attribute\CommercePaymentGateway;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayBase;
use Drupal\commerce_price\Price;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use GuzzleHttp\ClientInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Provides the GNU Taler payment gateway.
 */
#[CommercePaymentGateway(
  id: "commerce_gnu_taler",
  label: new TranslatableMarkup("GNU Taler"),
  display_label: new TranslatableMarkup("GNU Taler"),
  forms: [
    "offsite-payment" => PaymentForm::class,
  ],
  requires_billing_information: TRUE,
)]
class Taler extends OffsitePaymentGatewayBase implements TalerInterface, ContainerFactoryPluginInterface {

  /**
   * The HTTP client service.
   */
  protected ClientInterface $httpClient;

  /**
   * The order adapter.
   */
  protected OrderAdapter $orderAdapter;

  /**
   * The commerce mail handler.
   */
  protected MailHandlerInterface $mailHandler;

  /**
   * {@inheritdoc}
   *
   * @phpstan-param mixed[] $configuration
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->httpClient = $container->get(ClientInterface::class);
    $instance->orderAdapter = $container->get(OrderAdapter::class);
    $instance->mailHandler = $container->get(MailHandlerInterface::class);
    return $instance;
  }

  /**
   * {@inheritdoc}
   *
   * @phpstan-return mixed[]
   */
  public function defaultConfiguration(): array {
    return [
      'backend_url' => 'https://backend.demo.taler.net/instances/sandbox/',
      'backend_api_key' => 'secret-token:sandbox',
      'refund_delay' => 14,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   *
   * @phpstan-param mixed[] $form
   * @phpstan-return mixed[]
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form = parent::buildConfigurationForm($form, $form_state);
    $form['backend_url'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Backend URL'),
      '#maxlength' => 255,
      '#default_value' => $this->configuration['backend_url'],
      '#required' => TRUE,
      '#description' => $this->t('Set the URL of the Taler backend. (Example: https://backend.demo.taler.net/)'),
    ];
    $form['backend_api_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('API Key'),
      '#maxlength' => 255,
      '#default_value' => $this->configuration['backend_api_key'],
      '#required' => TRUE,
      '#description' => $this->t('Enter your API key to authenticate with the Taler backend. Will be sent as a "Bearer" token using the HTTP "Authorization" header. Typically should be prefixed with "secret-token:" (RFC 8959). This is not your password but a token to be obtained via the /token endpoint from the REST service.'),
    ];
    $form['refund_delay'] = [
      '#type' => 'number',
      '#title' => $this->t('Refund Delay'),
      '#default_value' => $this->configuration['refund_delay'],
      '#required' => TRUE,
      '#description' => $this->t('Set the number of days a customer has to request a refund'),
      '#min' => 1,
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   *
   * @phpstan-param mixed[] $form
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state): void {
    parent::validateConfigurationForm($form, $form_state);
    $values = $form_state->getValue($form['#parents']);
    assert(is_array($values));
    assert(is_string($values['backend_url']));
    assert(is_string($values['backend_api_key']));

    $client = new TalerClient($this->httpClient, $values['backend_url'], $values['backend_api_key']);
    if (!$client->verifyBackend()) {
      $form_state->setError($form['backend_url'], $this->t('Failed to contact taler backend'));
    }
  }

  /**
   * {@inheritdoc}
   *
   * @phpstan-param mixed[] $form
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    parent::submitConfigurationForm($form, $form_state);
    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);
      assert(is_array($values));
      assert(is_string($values['backend_url']));
      assert(is_string($values['backend_api_key']));

      $this->configuration['backend_url'] = $values['backend_url'];
      $this->configuration['backend_api_key'] = $values['backend_api_key'];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function onReturn(OrderInterface $order, Request $request): void {
    assert(is_string($this->configuration['backend_url']));
    assert(is_string($this->configuration['backend_api_key']));
    $client = new TalerClient($this->httpClient, $this->configuration['backend_url'], $this->configuration['backend_api_key']);

    $postOrderResponse = $order->getData('taler_post_order_response');
    if (!is_array($postOrderResponse) || !isset($postOrderResponse['order_id'])) {
      throw new PaymentGatewayException('Remote order data missing for this Taler transaction.');
    }
    assert(is_string($postOrderResponse['order_id']));

    $orderStatusResponse = $client->getOrderStatus($postOrderResponse['order_id']);
    if ($orderStatusResponse === FALSE) {
      throw new PaymentGatewayException('Unable to get remote order status.');
    }

    if ($orderStatusResponse['order_status'] !== 'paid') {
      throw new PaymentGatewayException(sprintf('Unexpected remote order status %s.', $orderStatusResponse['order_status']));
    }

    assert(isset($orderStatusResponse['contract_terms']));
    [$remoteCurrency, $remoteAmount] = explode(':', $orderStatusResponse['contract_terms']['amount'], 2);

    $paymentStorage = $this->entityTypeManager->getStorage('commerce_payment');
    $payment = $paymentStorage->create([
      'state' => 'completed',
      'amount' => new Price($remoteAmount, $this->orderAdapter->toSystemCurrency($remoteCurrency)),
      'payment_gateway' => $this->parentEntity->id(),
      'order_id' => $order->id(),
      'remote_id' => $postOrderResponse['order_id'],
      'remote_state' => $orderStatusResponse['order_status'],
    ]);
    $payment->save();
  }

  /**
   * {@inheritdoc}
   */
  public function createRedirectOrder(PaymentInterface $payment, string $returnUrl, string $cancelUrl): array {
    assert(is_string($this->configuration['backend_url']));
    assert(is_string($this->configuration['backend_api_key']));
    $client = new TalerClient($this->httpClient, $this->configuration['backend_url'], $this->configuration['backend_api_key']);

    $currencyCode = $payment->getAmount()?->getCurrencyCode();
    assert(is_string($currencyCode));
    $remoteCurrency = $this->orderAdapter->toRemoteCurrency($currencyCode);
    if (!$client->verifyCurrency($remoteCurrency)) {
      throw new PaymentGatewayException(sprintf('Currency %currency not supported by GNU Taler backend', $currencyCode));
    }

    $orderData = $this->orderAdapter->adaptCreateRedirectParams($payment, $returnUrl, $cancelUrl);

    // POST order data to taler backend.
    $postOrderResponse = $client->createRedirectOrder($orderData);
    if ($postOrderResponse === FALSE) {
      throw new PaymentGatewayException('Failed to create new order on GNU Taler backend');
    }

    // Save PostOrderResponse in commerce order.
    $order = $payment->getOrder();
    assert($order !== NULL);
    $order->setData('taler_post_order_response', $postOrderResponse)->save();

    // Extract token from response and construct redirect URL.
    ['order_id' => $remoteOrderId, 'token' => $remoteOrderToken] = $postOrderResponse;
    return [
      'url' => $this->configuration['backend_url'] . 'orders/' . $remoteOrderId,
      'data' => [
        'token' => $remoteOrderToken,
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function refundPayment(PaymentInterface $payment, ?Price $amount = NULL): void {
    assert(is_string($this->configuration['backend_url']));
    assert(is_string($this->configuration['backend_api_key']));
    $client = new TalerClient($this->httpClient, $this->configuration['backend_url'], $this->configuration['backend_api_key']);

    $this->assertPaymentState($payment, ['completed', 'partially_refunded']);

    $entireAmount = $payment->getAmount();
    assert($entireAmount !== NULL);

    // If not specified, refund the entire amount.
    $amount = $amount ?: $entireAmount;
    $this->assertRefundAmount($payment, $amount);

    $oldRefundAmount = $payment->getRefundedAmount();
    $newRefundAmount = $oldRefundAmount ? $oldRefundAmount->add($amount) : $amount;
    if ($newRefundAmount->lessThan($entireAmount)) {
      $payment->setState('partially_refunded');
    }
    else {
      $payment->setState('refunded');
    }

    $remoteId = $payment->getRemoteId();
    $refundRequest = $this->orderAdapter->adaptRefundRequest($payment, $newRefundAmount);
    $merchantRefundResponse = $client->postOrderRefund($remoteId, $refundRequest);
    if ($merchantRefundResponse === FALSE) {
      throw new PaymentGatewayException('Failed to create refund request on GNU Taler backend');
    }

    $payment->setRefundedAmount($newRefundAmount)->save();

    $uri = $merchantRefundResponse['taler_refund_uri'];

    $order = $payment->getOrder();
    assert($order !== NULL);
    $order->setData('taler_refund_uri', $uri)->save();

    // Send refund email.
    $to = $order->getEmail();

    $body = [
      '#theme' => 'commerce_gnu_taler_refund',
      '#order_entity' => $order,
      '#payment_entity' => $payment,
      '#refund_url' => $uri,
    ];

    $subject = $this->t('Order #@number refund granted', ['@number' => $order->getOrderNumber()]);
    $params = [
      'id' => 'gnu_taler_refund',
      'from' => $order->getStore()?->getEmailFromHeader(),
      'order' => $order,
    ];

    $this->mailHandler->sendMail($to, $subject, $body, $params);
  }

}
